"""
Scalar ODE example:
Vector and application class for Dahlquist's test problem
"""

import numpy as np

from pymgrit.core.application import Application
from pymgrit.core.vector import Vector


class VectorDahlquist(Vector):
    """
    Vector class for the Dahlquist test equation
    """

    def __init__(self, value):
        super(VectorDahlquist, self).__init__()
        self.value = value

    def __add__(self, other):
        tmp = VectorDahlquist(0)
        tmp.set_values(self.get_values() + other.get_values())
        return tmp

    def __sub__(self, other):
        tmp = VectorDahlquist(0)
        tmp.set_values(self.get_values() - other.get_values())
        return tmp

    def norm(self):
        return np.linalg.norm(self.value)

    def clone(self):
        return VectorDahlquist(self.value)

    def clone_zero(self):
        return VectorDahlquist(0)

    def clone_rand(self):
        return VectorDahlquist(np.random.rand(1)[0])

    def set_values(self, value):
        self.value = value

    def get_values(self):
        return self.value

    def pack(self):
        return self.value

    def unpack(self, value):
        self.value = value


class Dahlquist(Application):
    """
    Application class for Dahlquist's test problem,
        u' = lambda u,
    with lambda = -1 and IC u(0) = 1
    """

    def __init__(self, constant_lambda=-1, method='BE', *args, **kwargs):
        """
        Initialize Dahlquist application object
        :param method: method for solving Dahlquist's equation:
                       'BE' -> Backward Euler (default)
                       'FE' -> Forwared Euler
                       'TR' -> Trapezoidal rule
                       'MR' -> Implicit Mid-point Rule
        """
        super(Dahlquist, self).__init__(*args, **kwargs)
        self.vector_template = VectorDahlquist(0)  # Set the class to be used for each time point
        self.vector_t_start = VectorDahlquist(1)  # Set the initial condition
        self.lambda_value = constant_lambda
        if method == 'BE' or method == 'FE' or method == 'TR' or method == 'MR':
            self.method = method
        else:
            raise Exception(
                'Unknown method. Choose BE (Backward Euler), FE (Forward Euler), TR (Trapezoidal rule) ' +
                'or MR (implicit mid-point rule)')

    def step(self, u_start: VectorDahlquist, t_start: float, t_stop: float) -> VectorDahlquist:
        """
        Time integration routine for Dahlquist's test problem:
            BE: Backward Euler
            FE: Forward Euler
            TR: Trapezoidal rule
            MR: implicit Mid-point Rule

        :param u_start: approximate solution for the input time t_start
        :param t_start: time associated with the input approximate solution u_start
        :param t_stop: time to evolve the input approximate solution to
        :return: approximate solution for the input time t_stop
        """
        z = (t_stop - t_start) * self.lambda_value
        if self.method == 'BE':
            tmp = 1 / (1 - z) * u_start.get_values()
        elif self.method == 'FE':
            tmp = (1 + z) * u_start.get_values()
        elif self.method == 'TR':
            tmp = (1 + z / 2) / (1 - z / 2) * u_start.get_values()
        elif self.method == 'MR':
            k1 = -1 / (1 - z / 2) * u_start.get_values()
            tmp = u_start.get_values() + (t_stop - t_start) * k1
        return VectorDahlquist(tmp)



"""
Vector and application class for the 2D heat equation

Time integration methods:
   Forward Euler
   Backward Euler
   Crank-Nicolson
"""

from typing import Callable, Union
import numpy as np
from scipy.sparse import diags
from scipy.sparse import identity
from scipy.sparse.linalg import spsolve

from pymgrit.core.application import Application
from pymgrit.core.vector import Vector


class VectorHeat2D(Vector):
    """
    Vector class for the 2D heat equation
    """

    def __init__(self, nx, ny):
        """
        Constructor.

        :param nx: number of degrees of freedom in x-direction
        :param ny: number of degrees of freedom in y-direction
        """
        super(VectorHeat2D, self).__init__()
        self.nx = nx
        self.ny = ny
        self.values = np.zeros((self.nx, self.ny))

    def __add__(self, other):
        """
        Addition of two vector objects (self and other)

        :param other: vector object to be added to self
        :return: sum of vector object self and input object other
        """
        tmp = VectorHeat2D(self.nx, self.ny)
        tmp.set_values(self.get_values() + other.get_values())
        return tmp

    def __sub__(self, other):
        """
        Subtraction of two vector objects (self and other)

        :param other: vector object to be subtracted from self
        :return: difference of vector object self and input object other
        """
        tmp = VectorHeat2D(self.nx, self.ny)
        tmp.set_values(self.get_values() - other.get_values())
        return tmp

    def norm(self):
        """
        Norm of a vector object

        :return: 2-norm of vector object
        """
        return np.linalg.norm(self.values)

    def clone(self):
        """
        Clone vector object

        :rtype: vector object with zero values
        """
        tmp = VectorHeat2D(self.nx, self.ny)
        tmp.set_values(self.get_values())
        return tmp

    def clone_zero(self):
        """
        Initialize vector object with zeros

        :rtype: vector object with zero values
        """
        return VectorHeat2D(self.nx, self.ny)

    def clone_rand(self):
        """
        Initialize vector object with random values

        :rtype: vector object with random values
        """
        tmp = VectorHeat2D(self.nx, self.ny)
        tmp.set_values(np.random.rand(self.nx, self.ny))
        return tmp

    def set_values(self, values):
        """
        Set vector data

        :param values: values for vector object
        """
        self.values = values

    def get_values(self):
        """
        Get vector data

        :return: values of vector object
        """
        return self.values

    def pack(self):
        """
        Pack data

        :return: values of vector object
        """
        return self.values

    def unpack(self, values):
        """
        Unpack and set data

        :param values: values for vector object
        """
        self.values = values


class Heat2D(Application):
    """
    Application class for the heat equation in 2D space,
        u_t - a(u_xx + u_yy) = b(x,y,t),  a > 0,
        in [x_start, x_end] x [y_start, y_end] x (t_start, t_end],
    with homogeneous Dirichlet boundary conditions in space.
    """

    def __init__(self, x_start: float, x_end: float, y_start: float, y_end: float, nx: int, ny: int,
                 a: float, rhs: Callable = lambda x, y, t: 0 * x * y,
                 init_cond: Callable = lambda x, y: x * y * 0, method: str = 'BE',
                 bc_left: Union[int, float, Callable] = 0, bc_right: Union[int, float, Callable] = 0,
                 bc_bottom: Union[int, float, Callable] = 0, bc_top: Union[int, float, Callable] = 0, *args, **kwargs):
        """
        Constructor.

        :param x_start: left bound of x-domain
        :param x_end: right bound of x-domain
        :param y_start: left bound of y-domain
        :param y_end: right bound of y-domain
        :param nx: number of points in x-direction
        :param ny: number of points in y-direction
        :param a: thermal conductivity
        :param rhs: right-hand side
        :param init_cond: initial condition
        :param method: method for solving Dahlquist's equation:
                       'BE' -> Backward Euler (default)
                       'FE' -> Forwared Euler
                       'CN' -> Crank-Nicolson
        :param bc_left: boundary condition for left boundary
        :param bc_right: boundary condition for right boundary
        :param bc_bottom: boundary condition for bottom boundary
        :param bc_top: boundary condition for top boundary
        """
        super(Heat2D, self).__init__(*args, **kwargs)
        # Spatial domain (including boundary points)
        self.x_start = x_start
        self.x_end = x_end
        self.y_start = y_start
        self.y_end = y_end
        self.x = np.linspace(x_start, x_end, nx)
        self.y = np.linspace(y_start, y_end, ny)
        self.x_2d = self.x[:, np.newaxis]
        self.y_2d = self.y[np.newaxis, :]
        self.nx = nx
        self.ny = ny
        self.dx = self.x[1] - self.x[0]
        self.dy = self.y[1] - self.y[0]

        # Thermal conductivity
        self.a = a

        # Set right-hand side routine
        self.rhs = rhs

        # Set time integration method
        if method == 'BE':
            self.theta = 1
        elif method == 'FE':
            self.theta = 0
        elif method == 'CN':
            self.theta = 0.5
        else:
            raise Exception("Unknown method. Choose BE (Backward Euler), FE (Forward Euler) or CN (Crank-Nicolson")

        # Set Dirichlet boundary conditions
        if isinstance(bc_left, (float, int)):
            self.bc_left: Callable = lambda y: bc_left
        elif callable(bc_left):
            self.bc_left: Callable = bc_left
        else:
            raise Exception("Choose float, int or function for boundary condition bc_left")

        if isinstance(bc_right, (float, int)):
            self.bc_right: Callable = lambda y: bc_right
        elif callable(bc_right):
            self.bc_right: Callable = bc_right
        else:
            raise Exception("Choose float, int or function for boundary condition bc_right")

        if isinstance(bc_bottom, (float, int)):
            self.bc_bottom: Callable = lambda x: bc_bottom
        elif callable(bc_bottom):
            self.bc_bottom: Callable = bc_bottom
        else:
            raise Exception("Choose float, int or function for boundary condition bc_bottom")

        if isinstance(bc_top, (float, int)):
            self.bc_top: Callable = lambda x: bc_top
        elif callable(bc_top):
            self.bc_top: Callable = bc_top
        else:
            raise Exception("Choose float, int or function for boundary condition bc_top")

        # Set (spatial) identity matrix and spatial discretization matrix
        self.space_disc = self.compute_matrix()
        self.identity = identity(self.nx * self.ny, dtype='float', format='csr')

        # Set the data structure for any user-defined time point
        self.vector_template = VectorHeat2D(self.nx, self.ny)

        # Set initial condition
        self.init_cond = init_cond
        self.vector_t_start = VectorHeat2D(self.nx, self.ny)
        init = self.init_cond(self.x_2d, self.y_2d)
        init[:, 0] = self.bc_left(self.x)
        init[:, -1] = self.bc_right(self.x)
        init[-1, :] = self.bc_bottom(self.y)
        init[0, :] = self.bc_top(self.y)
        self.vector_t_start.set_values(init)

    def compute_matrix(self):
        """
        Define spatial discretization matrix for 2D heat equation

        Second-order central finite differences with matrix stencil
            [          -f_y          ]
            [-f_x  2(f_x + f_y)  -f_x]
            [          -f_y          ]
        with f_x = (a / dx^2) and f_y = (a / dy^2)
        """
        fx = self.a / self.dx ** 2
        fy = (self.a / self.dy ** 2)
        n = self.nx * self.ny

        main = np.ones(n) * (2 * (fx + fy))
        main[:self.ny] = main[-self.ny:] = 0
        main[self.ny::self.ny] = main[2 * self.ny - 1::self.ny] = 0

        upper = np.ones(n - 1) * (-fy)
        upper[:self.ny] = upper[-self.ny:] = 0
        upper[self.ny::self.ny] = upper[2 * self.ny - 1:: self.ny] = 0

        lower = np.ones(n - 1) * (-fy)
        lower[:self.ny - 1] = lower[-self.ny:] = 0
        lower[self.ny - 1::self.ny] = lower[2 * self.ny - 2:: self.ny] = 0

        upper2 = np.ones(n - self.ny) * (-fx)
        upper2[:self.ny] = 0
        upper2[self.ny::self.ny] = upper2[2 * self.ny - 1:: self.ny] = 0

        lower2 = np.ones(n - self.ny) * (-fx)
        lower2[-self.ny:] = 0
        lower2[::self.ny] = lower2[self.ny - 1::self.ny] = 0

        matrix = diags(diagonals=[main, lower, upper, lower2, upper2],
                       offsets=[0, -1, 1, -self.ny, self.ny],
                       shape=(n, n), format='csr')
        return matrix

    def compute_rhs(self, u_start: np.ndarray, t_start: float, t_stop: float):
        """
        Right-hand side of spatial system for implicit time integration methods

        :param u_start: approximate solution for the input time t_start
        :param t_start: time associated with the input approximate solution u_start
        :param t_stop: time to evolve the input approximate solution to
        :return: right-hand side of spatial system at each time step in case of implicit time integration
        """

        b = np.zeros((self.nx, self.ny))

        if self.theta == 1:
            # BE
            b[1:-1, 1:-1] = u_start[1:-1, 1:-1] + \
                            (t_stop - t_start) * self.rhs(x=self.x_2d[1:-1], y=self.y_2d[:, 1:-1], t=t_stop)
        else:
            # CN: theta = 1/2
            tmp = (self.identity - self.theta * (t_stop - t_start) * self.space_disc) * u_start.flatten()
            b += tmp.reshape(self.nx, self.ny)

            # add RHS of PDE
            b[1:-1, 1:-1] += self.theta * (t_stop - t_start) * self.rhs(x=self.x_2d[1:-1], y=self.y_2d[:, 1:-1],
                                                                        t=t_stop) + \
                             (1 - self.theta) * (t_stop - t_start) * self.rhs(x=self.x_2d[1:-1], y=self.y_2d[:, 1:-1],
                                                                              t=t_start)
        # Set boundary conditions
        b[:, 0] = self.bc_left(self.x)
        b[:, -1] = self.bc_right(self.x)
        b[-1, :] = self.bc_bottom(self.y)
        b[0, :] = self.bc_top(self.y)
        return b.flatten()

    def step(self, u_start: VectorHeat2D, t_start: float, t_stop: float) -> VectorHeat2D:
        """
        Time integration routine for 2D heat equation example problem:
            Backward Euler (BE)
            Forward Euler (FE)
            Crank-Nicolson (CN)

        One-step method
           (u_i - u_{i-1})/dt + (theta*L*u_i + (1-theta)*L*u_{i-1} = theta*b_i + (1-theta)*b_{i-1},
        where L = self.space_disc is the spatial discretization operator and
           theta =  0  -> FE
           theta = 1/2 -> CN
           theta =  1  -> BE

        :param u_start: approximate solution for the input time t_start
        :param t_start: time associated with the input approximate solution u_start
        :param t_stop: time to evolve the input approximate solution to
        :return: approximate solution at input time t_stop
        """
        if self.theta == 0:
            # FE
            old = u_start.get_values()
            new = np.zeros((self.nx, self.ny))

            # Set BCs
            new[:, 0] = self.bc_left(self.x)
            new[:, -1] = self.bc_right(self.x)
            new[-1, :] = self.bc_bottom(self.y)
            new[0, :] = self.bc_top(self.y)

            tmp = (self.identity - (t_stop - t_start) * self.space_disc) * old.flatten()
            new += tmp.reshape(self.nx, self.ny)

            # add RHS of PDE
            new[1:-1, 1:-1] += (t_stop - t_start) * self.rhs(x=self.x_2d[1:-1], y=self.y_2d[:, 1:-1], t=t_start)

            new.flatten()

        else:
            # set up and solve linear system
            rhs = self.compute_rhs(u_start=u_start.get_values(), t_start=t_start, t_stop=t_stop)
            new = spsolve((t_stop - t_start) * self.theta * self.space_disc + self.identity, rhs)
        ret = VectorHeat2D(self.nx, self.ny)
        ret.set_values(new.reshape(self.nx, self.ny))
        return ret



"""
MGRIT optimized for the induction
machine model "im_3kW". (https://gitlab.onelab.info/doc/models/-/wikis/Electric-machines)
Using joule losses as stopping criteria
"""

import logging
import time
import numpy as np

from pymgrit.core.mgrit import Mgrit


class MgritMachineConvJl(Mgrit):
    """
    MGRIT optimized for the getdp induction machine
    """

    def __init__(self, compute_f_after_convergence: bool, *args, **kwargs) -> None:
        """
        MGRIT optimized for the getdp induction machine

        :param compute_f_after_convergence: computes solution of F-points at the end
        """
        super(MgritMachineConvJl, self).__init__(*args, **kwargs)
        self.last_it = []
        self.compute_f_after_convergence = compute_f_after_convergence
        self.convergence_criteria(iteration=0)

    def nested_iteration(self) -> None:
        """
        Generates an initial approximation on the finest grid
        by solving the problem on the coarsest grid and interpolating
        the approximation to the finest level.

        Performs nested iterations with a sin rhs
        """
        change = False
        tmp_problem_pwm = np.zeros(len(self.problem))
        if self.problem[0].pwm:
            change = True
            for lvl in range(len(self.problem)):
                tmp_problem_pwm[lvl] = self.problem[lvl].pwm
                self.problem[lvl].fopt[-1] = 0

        self.forward_solve(self.lvl_max - 1)

        for lvl in range(self.lvl_max - 2, -1, -1):
            for i in range(len(self.index_local[lvl + 1])):
                self.u[lvl][self.index_local_c[lvl][i]] = self.interpolation[lvl](
                    u=self.u[lvl + 1][self.index_local[lvl + 1][i]])

            self.f_exchange(lvl)
            self.c_exchange(lvl)
            if lvl > 0:
                self.iteration(lvl, 'V', 0, True)

        if change:
            for lvl in range(len(self.problem)):
                self.problem[lvl].fopt[-1] = tmp_problem_pwm[lvl]

    def iteration(self, lvl: int, cycle_type: str, iteration: int, first_f: bool):
        """
        MGRIT iteration on level lvl

        :param lvl: MGRIT level
        :param cycle_type: Cycle type
        :param iteration: Number of current iteration
        :param first_f: F-relaxation at the beginning
        """
        if lvl == self.lvl_max - 1:
            self.forward_solve(lvl=lvl)
            return

        if first_f:
            self.f_relax(lvl=lvl)
            self.f_exchange(lvl=lvl)

        for _ in range(self.cf_iter):
            self.c_relax(lvl=lvl)
            self.c_exchange(lvl=lvl)
            self.f_relax(lvl=lvl)
            self.f_exchange(lvl=lvl)

        self.fas_residual(lvl=lvl)

        self.iteration(lvl=lvl + 1, cycle_type=cycle_type, iteration=iteration, first_f=True)

        self.error_correction(lvl=lvl)

        if lvl > 0:
            self.f_relax(lvl=lvl)

        if lvl != 0 and cycle_type == 'F':
            self.f_exchange(lvl=lvl)
            self.iteration(lvl=lvl, cycle_type='V', iteration=iteration, first_f=False)

    def convergence_criteria(self, iteration: int) -> None:
        """
        Maximum norm of all C-points

        :param iteration: iteration number
        """
        if len(self.last_it) != len(self.index_local_c[0]):
            self.last_it = np.zeros(len(self.index_local_c[0]))
        new = np.zeros_like(self.last_it)
        j = 0
        tmp = 0
        if self.index_local_c[0].size > 0:
            for i in np.nditer(self.index_local_c[0]):
                new[j] = self.u[0][i].jl
                j = j + 1
            tmp = 100 * np.max(
                np.abs(np.abs(np.divide((new - self.last_it), new, out=np.zeros_like(self.last_it), where=new != 0))))

        tmp = self.comm_time.allgather(tmp)
        self.conv[iteration] = np.max(np.abs(tmp))
        self.last_it = np.copy(new)

    def solve(self) -> dict:
        """
        Driver function for solving the problem using MGRIT.

        Performs MGRIT iterations until a stopping criterion is fulfilled or
        the maximum number of iterations is reached.

        Post operation for computing solution at F-points

        :return: dictionary with residual history, setup time, and solve time
        """
        tmp_output_fcn = self.output_fcn
        self.output_fcn = None
        super(MgritMachineConvJl, self).solve()
        self.output_fcn = tmp_output_fcn
        if self.compute_f_after_convergence:
            if self.comm_time.Get_rank() == 0:
                logging.info("Start post-processing: F-relax")
            runtime_pp_start = time.time()
            self.f_relax(lvl=0)
            runtime_pp_stop = time.time()
            if self.comm_time.Get_rank() == 0:
                logging.info(f"Post-processing took {runtime_pp_stop - runtime_pp_start} s")
        self.last_it = np.zeros_like(self.last_it)
        if self.output_fcn is not None:
            self.output_fcn(self)
        return {'conv': self.conv[np.where(self.conv != 0)], 'time_setup': self.runtime_setup,
                'time_solve': self.runtime_solve}



"""
Vector class for 1D heat problem

Note: values at two consecutive time points are grouped as pairs
"""

import numpy as np

from pymgrit.core.vector import Vector


class VectorHeat1D2Pts(Vector):
    """
    Vector class for grouping values at two consecutive time points
    """

    def __init__(self, size, dtau):
        """
        Constructor.
        One vector object contains values at two consecutive time points and spacing between these time points.

        :param size: number of spatial degrees of freedom
        :param dtau: time-step size within pair
        """
        super(VectorHeat1D2Pts, self).__init__()
        self.size = size
        self.dtau = dtau
        self.values_first_time_point = np.zeros(size)
        self.values_second_time_point = np.zeros(size)

    def __add__(self, other):
        """
        Addition of two vector objects (self and other)

        :param other: vector object to be added to self
        :return: sum of vector object self and input object other
        """
        tmp = VectorHeat1D2Pts(self.size, self.dtau)
        first_self, second_self, dtau_self = self.get_values()
        first_other, second_other, dtau_other = other.get_values()
        tmp.set_values(first_self + first_other, second_self + second_other, dtau_self)
        return tmp

    def __sub__(self, other):
        """
        Subtraction of two vector objects (self and other)

        :param other: vector object to be subtracted from self
        :return: difference of vector object self and input object other
        """
        tmp = VectorHeat1D2Pts(self.size, self.dtau)
        first_self, second_self, dtau_self = self.get_values()
        first_other, second_other, dtau_other = other.get_values()
        tmp.set_values(first_self - first_other, second_self - second_other, dtau_self)
        return tmp

    def norm(self):
        """
        Norm of a vector object

        :return: 2-norm of vector object
        """
        return np.linalg.norm(np.append(self.values_first_time_point, self.values_second_time_point))

    def clone(self):
        """
        Clone vector object

        :return: vector object with zero values
        """
        tmp = VectorHeat1D2Pts(self.size, self.dtau)
        tmp.set_values(self.values_first_time_point, self.values_second_time_point, self.dtau)
        return tmp

    def clone_zero(self):
        """
        Initialize vector object with zeros

        :return: vector object with zero values
        """
        return VectorHeat1D2Pts(self.size, self.dtau)

    def clone_rand(self):
        """
        Initialize vector object with random values

        :return: vector object with random values
        """
        tmp = VectorHeat1D2Pts(self.size, self.dtau)
        tmp.set_values(np.random.rand(self.size), np.random.rand(self.size), self.dtau)
        return tmp

    def get_values(self):
        """
        Get vector data

        :return: tuple of values of member variables
        """
        return self.values_first_time_point, self.values_second_time_point, self.dtau

    def set_values(self, first_time_point, second_time_point, dtau):
        """
        Set vector data

        :param first_time_point: values for first time point
        :param second_time_point: values for second time point
        :param dtau: time-step size within pair
        """
        self.values_first_time_point = first_time_point
        self.values_second_time_point = second_time_point
        self.dtau = dtau

    def pack(self):
        """
        Pack data

        :return: values of vector object
        """
        return np.array([self.values_first_time_point, self.values_second_time_point])

    def unpack(self, values):
        """
        Unpack and set data

        :param values: values for vector object
        """
        self.values_first_time_point = values[0]
        self.values_second_time_point = values[1]



"""
Grid Transfer class for the induction
machine model "im_3kW". (https://gitlab.onelab.info/doc/models/-/wikis/Electric-machines)

For more details see: https://arxiv.org/abs/1912.03106

Interpolation: standard finite element interpolation
Restriction: injection
"""

import copy

import numpy as np

from pymgrit.core.grid_transfer import GridTransfer
from pymgrit.induction_machine.vector_machine import VectorMachine
from pymgrit.induction_machine.helper import check_version, compute_data, interpolation_factors, compute_mesh_transfer


class GridTransferMachine(GridTransfer):
    """
    Grid Transfer class for the induction machine im_3kW
    Interpolation: standard finite element interpolation
    Restriction: injection
    """

    def __init__(self, coarse_grid, fine_grid, path_meshes):
        """
        Constructor. Compute the transfer grid between both given grids in both directions.

        :param coarse_grid: Coarse grid name
        :param fine_grid:  Fine grid name
        :param path_meshes: Path to meshes
        """
        super(GridTransferMachine, self).__init__()
        data_coarse_pre = path_meshes + coarse_grid + '.pre'
        data_coarse_msh = path_meshes + coarse_grid + '.msh'
        check_version(msh_file=data_coarse_msh)
        data_coarse = compute_data(data_coarse_pre, data_coarse_msh, 0)

        data_fine_pre = path_meshes + fine_grid + '.pre'
        data_fine_msh = path_meshes + fine_grid + '.msh'
        check_version(msh_file=data_fine_msh)
        data_fine = compute_data(data_fine_pre, data_fine_msh, len(data_coarse['corToUn']))

        self.transfer_data = interpolation_factors(data_coarse=data_coarse, data_fine=data_fine)

    def restriction(self, u: VectorMachine) -> VectorMachine:
        """
        Restriction

        :param u: approximate solution vector
        :return: input solution vector u restricted to a coarse grid
        """
        ret = copy.deepcopy(u)
        ret.u_middle = ret.u_middle[:self.transfer_data['sizeLvlStart']]
        ret.u_middle_size = self.transfer_data['sizeLvlStart']
        return ret

    def interpolation(self, u: VectorMachine) -> VectorMachine:
        """
        Interpolation

        :param u: approximate solution vector
        :return: input solution vector u interpolated to a fine grid
        """
        ret = copy.deepcopy(u)

        new_middle = np.zeros(self.transfer_data['sizeLvlStop'] - self.transfer_data['sizeLvlStart'])

        new_u_inner = compute_mesh_transfer(
            u.u_middle[self.transfer_data['mappingInner']], self.transfer_data['vtxInner'],
            self.transfer_data['wtsInner'], self.transfer_data['addBoundInner'], 0)

        new_u_outer = compute_mesh_transfer(
            u.u_middle[self.transfer_data['mappingOuter']], self.transfer_data['vtxOuter'],
            self.transfer_data['wtsOuter'], self.transfer_data['addBoundOuter'], 0)
        new_middle[:len(u.u_middle)] = u.u_middle
        new_middle[self.transfer_data['mappingInnerNew']] = new_u_inner
        new_middle[self.transfer_data['mappingOuterNew']] = new_u_outer
        ret.u_middle = np.append(ret.u_middle, new_middle)
        ret.u_middle_size = len(ret.u_middle)
        return ret



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import FancyBboxPatch
import matplotlib.colors as mcolors

from pymgrit.core.mgrit import Mgrit


class MgritWithPlots(Mgrit):
    def __init__(self, *args, **kwargs) -> None:
        """
        Cumstomized MGRIT constructor
        :param args:
        :param kwargs:
        """
        # Call parent constructor
        super(MgritWithPlots, self).__init__(*args, **kwargs)

    def plot(self, plot_function):
        if self.comm_time.rank == 0:
            if self.spatial_parallel:
                if self.comm_space == 0:
                    plot_function()
            else:
                plot_function()

    def plot_convergence(self, save_name=None, fig_size_x=6.4, fig_size_y=4.8, dpi=100, text_size=15):
        def plot_conv():
            fig = plt.figure(figsize=(fig_size_x, fig_size_y), dpi=dpi, constrained_layout=True)
            fig.set_constrained_layout_pads(w_pad=0.1, h_pad=0)
            iters = np.arange(1, len(self.conv[np.where(self.conv != 0)]) + 1)
            plt.semilogy(iters, self.conv[np.where(self.conv != 0)], 'o-', lw=int(text_size / 2), markersize=text_size)
            plt.xticks(iters)
            plt.xlabel('iter #', size=text_size, weight='bold')
            plt.ylabel('residual norm', size=text_size, weight='bold')
            plt.yticks(size=text_size, weight='bold')
            plt.xticks(size=text_size, weight='bold')
            if save_name is not None:
                plt.savefig(save_name)
            plt.show()

        self.plot(plot_function=plot_conv)

    def plot_parallel_distribution(self, time_procs=None, save_name=None, fig_size_x=6.4, fig_size_y=4.8, dpi=100,
                                   text_size=15):

        if time_procs is None:
            time_procs = self.comm_time_size

        def plot_parallel_dist():
            fig = plt.figure(figsize=(fig_size_x, fig_size_y), dpi=dpi, constrained_layout=True)
            fig.set_constrained_layout_pads(w_pad=0.1, h_pad=0)
            ax = plt.gca()

            size_lvl = []
            left_lower = []
            boxes = []

            colors = list(mcolors.BASE_COLORS)
            if 'k' in colors: colors.remove('k')

            if time_procs > len(colors):
                for i in range(time_procs - len(colors)):
                    colors.append(list(np.random.rand(3)))

            nt = len(self.problem[0].t)

            split = self.split_into(number_points=nt, number_processes=time_procs)

            for i in range(len(split)):
                left_lower.append([np.sum(split[:i]), -0.2])

            for i in range(self.lvl_max):
                for j in range(len(split)):
                    if i != 0:
                        boxes.append(FancyBboxPatch(
                            [a + b for a, b in zip(left_lower[j], [-0.25, i])], split[j] - 0.25, 0.4,
                            boxstyle="round,pad=-0.0040,rounding_size=0.015", fc=colors[j],
                            label='Process ' + str(j)))
                    else:
                        boxes.append(FancyBboxPatch(
                            [a + b for a, b in zip(left_lower[j], [-0.25, i])], np.sum(split) - 0.25, 0.4,
                            boxstyle="round,pad=-0.0040,rounding_size=0.015", fc=colors[j],
                            label='Process ' + str(j)))
                        break
            for item in boxes:
                ax.add_patch(item)

                legend = plt.legend(handles=boxes[-time_procs:], fontsize=text_size, loc='center left',
                                    bbox_to_anchor=(1, 0.5), prop={'weight': 'bold', 'size': text_size})

            for i in range(self.lvl_max):
                if i != self.lvl_max -1:
                    tmp_coarse = np.where(np.in1d(self.problem[0].t, self.problem[i + 1].t))[0]
                tmp_fine = np.where(np.in1d(self.problem[0].t, self.problem[i].t))[0]
                plt.plot([0, nt - 1], [i, i], color='k')

                for j in range(nt):
                    if j in tmp_coarse:
                        plt.plot([j, j],[((self.lvl_max - 1) - i) + 0.1, ((self.lvl_max - 1) - i) - 0.1], color='k')
                    elif j in tmp_fine:
                        plt.plot([j, j],[((self.lvl_max - 1) - i) + 0.05, ((self.lvl_max - 1) - i) - 0.05],color='k')

            [s.set_visible(False) for s in ax.spines.values()]
            ax.set_xticks([])
            ax.set_yticks(np.arange(0, self.lvl_max, 1).tolist())
            ax.set_yticklabels(['Level ' + str(i) for i in range(self.lvl_max - 1, -1, -1)])
            plt.yticks(size=text_size, weight='bold')
            if save_name is not None:
                plt.savefig(save_name, dpi=dpi)
            plt.show()

        self.plot(plot_function=plot_parallel_dist)

    def plot_cycle(self, iterations=1, save_name=None, fig_size_x=6.4, fig_size_y=4.8, dpi=100, text_size=15):

        def plot_cyc():
            def mask(len, result):
                strings = np.array([''] * len, dtype=object)
                strings[np.where(result < 0)[0]] = 'F' + self.cf_iter * 'CF'
                strings[np.where(result == 0)[0]] = 'S'
                strings[np.where(result > 0)[0]] = 'F'
                return strings

            def plot_part(iter, len_res, len_nes, result, nest_it=False):
                if nest_it:
                    x_axis = np.arange(0, len_nes, 1)
                    strings = mask(len=len_nes, result=result)
                    strings[-1] = ''
                else:
                    x_axis = np.arange(iter * len_res + len_nes, len_nes + (iter + 1) * len_res, 1)
                    strings = mask(len=len_res, result=result)
                if iter < iterations - 1 or nest_it:
                    plt.axvline(x=x_axis[-1] + 0.4, color='k', linestyle='--')
                plt.plot(x_axis, np.abs(result), color='k')
                if iter != 0:
                    strings[0] = strings[0][1:]
                for i in range(len(result)):
                    plt.text(x_axis[i], np.abs(result[i]), strings[i], size=text_size, weight='bold', ha="center",
                             va="center",
                             bbox=dict(boxstyle="round", facecolor='white', edgecolor='k'))
                return np.average(x_axis)

            fig = plt.figure(figsize=(fig_size_x, fig_size_y), dpi=dpi, constrained_layout=True)
            fig.set_constrained_layout_pads(w_pad=0.1, h_pad=0)
            ax = plt.gca()

            result_nes = result_iter = np.zeros(0)
            x_ticks = []
            if self.nes_it:
                for i in range(self.lvl_max-1):
                    result_nes = np.hstack((result_nes, np.linspace(-i, i, 2 * i + 1)))
                result_nes = np.hstack((result_nes, self.lvl_max-1))
                x_ticks.append(plot_part(iter=0, len_res=0, len_nes=len(result_nes), result=result_nes, nest_it=True))
            if self.cycle_type == 'V':
                result_iter = np.hstack(
                    (result_iter, np.linspace(-self.lvl_max + 1, self.lvl_max - 1, 2 * (self.lvl_max - 1) + 1)))
            elif self.cycle_type == 'F':
                result_iter = np.hstack((result_iter, np.linspace(-self.lvl_max + 1, 0, self.lvl_max)))[:-1]
                for i in range(self.lvl_max):
                    result_iter = np.hstack((result_iter, np.linspace(-i, i, 2 * i + 1)))
            for iter in range(0, iterations):
                x_ticks.append(
                    plot_part(iter=iter, len_res=len(result_iter), len_nes=len(result_nes), result=result_iter,
                              nest_it=False))

            [s.set_visible(False) for s in ax.spines.values()]
            ax.set_xticks(x_ticks)
            ax.set_xticklabels(
                ['Nested iteration'] + ['Iteration ' + str(i+1) for i in range(iterations)] if self.nes_it else [
                    'Iteration ' + str(i+1) for i in range(iterations)])
            ax.set_yticks(np.arange(0, self.lvl_max, 1).tolist())
            ax.set_yticklabels(['Level ' + str(i) for i in range(self.lvl_max - 1, -1, -1)])
            # ax.tick_params(axis='both', which='both', length=0)
            plt.yticks(size=text_size, weight='bold')
            plt.xticks(size=text_size, weight='bold')
            plt.xlabel(' ', size=text_size, weight='bold')
            plt.ylabel(' ', size=text_size, weight='bold')
            if save_name is not None:
                plt.savefig(save_name)
            plt.show()

        self.plot(plot_function=plot_cyc)



"""
MGRIT solver in FAS formulation
"""
import time
import logging
import sys
from operator import itemgetter
from typing import Tuple, List
from itertools import groupby
import numpy as np

from mpi4py import MPI

from pymgrit.core.application import Application
from pymgrit.core.grid_transfer import GridTransfer
from pymgrit.core.grid_transfer_copy import GridTransferCopy


class Mgrit:
    """
    MGRIT solver class

    Implementation of MGRIT FAS algorithm for solving
    time-stepping problems of the form
      u_i = Phi(u_{i-1}),
    where Phi propagates u_{i-1} from t = t_{i-1} to t = t_i.

    It is assumed that the problems have a constant dimension
    and the solved space-time matrix stencil is [-Phi I].
    """

    def __init__(self, problem: List[Application], transfer: List[GridTransfer] = None,
                 max_iter: int = 100, tol: float = 1e-7, nested_iteration: bool = True, cf_iter: int = 1,
                 cycle_type: str = 'V', comm_time: MPI.Comm = MPI.COMM_WORLD, comm_space: MPI.Comm = MPI.COMM_NULL,
                 logging_lvl: int = logging.INFO, output_fcn=None, output_lvl=1, t_norm=2,
                 random_init_guess: bool = False) -> None:
        """
        Initialize MGRIT solver.

        :param problem: List of problems (one for each MGRIT level)
        :param transfer: List of spatial transfer operators (one for each pair of consecutive MGRIT levels)
        :param max_iter: Maximum number of iterations
        :param tol: stopping tolerance
        :param nested_iteration: With (True) or without (False) nested iterations
        :param cf_iter: Number of CF relaxations in each MGRIT iteration
        :param cycle_type: 'F' or 'V' cycle
        :param comm_time: Time communicator
        :param comm_space: Space communicator
        :param logging_lvl: Logging level:
               Value = 10: Debug logging level -> Runtime of all components
               Value = 20: Info logging level  -> Information per MGRIT iteration + summary at the end
               Value = 30: No logging level    -> No information
        :param output_fcn: Function for saving solution values to file
        :param output_lvl: Output level, possible values 0, 1, 2:
               0 -> output_fcn is never called
               1 -> output_fcn is called at the end of the simulation
               2 -> output_fcn is called after each MGRIT iteration
        :param random_init_guess: Use (True) or do not use (False) random initial guess
        """
        logging.basicConfig(format='%(levelname)s - %(asctime)s - %(message)s', datefmt='%d-%m-%y %H:%M:%S',
                            level=logging_lvl, stream=sys.stdout)

        # Set standard grid transfer operators if no transfer operators are given
        if transfer is None:
            transfer = [GridTransferCopy() for _ in range(len(problem) - 1)]

        # Check input parameters
        if len(problem) != (len(transfer) + 1):
            raise Exception('There should be exactly one transfer operator for each level except the coarsest grid')

        for i in range(len(problem) - 1):
            if len(problem[i].t) < len(problem[i + 1].t):
                raise Exception(
                    'The time grid on level ' + str(i + 1) + ' contains more time points than level ' + str(i))

        if cycle_type != 'V' and cycle_type != 'F':
            raise Exception("Cycle-type " + str(cycle_type) + " is not implemented. Choose 'V' or 'F'")

        if output_lvl not in [0, 1, 2]:
            raise Exception("Unknown output level. Choose 0, 1 or 2.")

        for lvl in range(1, len(problem)):
            if len(set(problem[lvl - 1].t.tolist()).intersection(set(problem[lvl].t.tolist()))) != len(problem[lvl].t):
                raise Exception(
                    'Some points from level ' + str(lvl - 1) + ' are not points of level ' + str(lvl))

        if t_norm not in [1, 2, 3]:
            raise Exception(
                'Unknown norm. Please choose 1 (one norm), 2 (two-norm) or 3 (inf-norm)')

        self.comm_time = comm_time
        self.comm_space = comm_space
        self.comm_time_rank = self.comm_time.Get_rank()
        self.comm_time_size = self.comm_time.Get_size()

        if self.comm_time_size > len(problem[0].t):
            raise Exception('More processors than time points. Not useful and not implemented yet')

        # Check if spatial parallelism is used
        if self.comm_space != MPI.COMM_NULL:
            self.spatial_parallel = True
            self.comm_space_rank = self.comm_space.Get_rank()
            self.comm_space_size = self.comm_space.Get_size()
        else:
            self.spatial_parallel = False
            self.comm_space_rank = -99
            self.comm_space_size = 1

        # Start timer for setup time
        self.comm_time.barrier()
        runtime_setup_start = time.time()
        self.log_info(f"Start setup")

        # Initialize MGRIT parameters
        self.problem = problem  # List of problems (one per MGRIT level)
        self.lvl_max = len(problem)  # Max number of MGRIT levels
        self.step = []  # List of time integration routines (one per MGRIT level)
        self.u = []  # List of solutions (one per MGRIT level)
        self.v = []  # List of restricted unknowns (one per MGRIT level)
        self.g = []  # List of FAS right-hand sides (one per MGRIT level)
        self.t = []  # List of local time intervals (one per MGRIT level)
        self.m = []  # List of coarsening factors
        self.restriction = []  # List of restriction operators (one per MGRIT level - except for coarsest)
        self.interpolation = []  # List of interpolation operators (one per MGRIT level - except for coarsest)
        self.tol = tol  # Convergence tolerance
        self.conv = np.zeros(max_iter + 1)  # Convergence information after each iteration
        self.cf_iter = cf_iter  # Number of CF-relaxations
        self.cycle_type = cycle_type  # Cycle type, F or V
        self.random_init_guess = random_init_guess  # Random initial guess
        self.iter_max = max_iter  # Maximum number of iterations
        self.solve_iter = 0  # MGRIT iteration number; for output
        self.nes_it = nested_iteration  # Local nested iteration value
        self.runtime_solve = 0  # Solve runtime
        self.runtime_setup = 0  # Setup runtime
        self.int_start = 0  # Index of first time point of local time interval
        self.int_stop = 0  # Index of last time points of local time interval
        self.cpts = []  # Global index of local C-points
        self.index_local_c = []  # Local indices of C-Points
        self.index_local_f = []  # Local indices of F-Points
        self.index_local = []  # Local indices of all points
        self.g_coarsest = []  # FAS residual for the time stepping on coarsest grid
        self.u_coarsest = []  # Solution for the time stepping on coarsest grid
        self.comm_front = []  # Communication inside F-relax per MGRIT level
        self.comm_back = []  # Communication inside F-relax per MGRIT level
        self.first_is_f_point = []  # Communication after C-relax
        self.first_is_c_point = []  # Communication after F-relax
        self.last_is_f_point = []  # Communication after F-relax
        self.last_is_c_point = []  # Communication after C-relax
        self.send_to = []  # Which process contains next time point
        self.get_from = []  # Which process contains previous time point
        self.global_t = []  # Global time information
        self.t_norm = 1 if t_norm == 1 else None if t_norm == 2 else np.inf  # Time norm

        # Set output level and output function
        self.output_lvl = output_lvl  # Output level, only 0,1,2
        if output_fcn is not None and callable(output_fcn):
            self.output_fcn = output_fcn
        else:
            self.output_fcn = None

        # Set local MGRIT parameters
        for lvl in range(self.lvl_max):
            self.t.append(np.copy(problem[lvl].t))
            if lvl != self.lvl_max - 1:
                self.restriction.append(transfer[lvl].restriction)
                self.interpolation.append(transfer[lvl].interpolation)
            if lvl < self.lvl_max - 1:
                tmp_cpts = np.where(np.in1d(self.problem[lvl].t, self.problem[lvl + 1].t))[0]
                tmp_m = np.mean(np.absolute(
                    tmp_cpts[1:] if len(tmp_cpts[1:]) > 0 else [1] - tmp_cpts[:-1] if len(tmp_cpts[:-1]) > 0 else [1]),
                                0)
                self.m.append(int(tmp_m))
                if not tmp_m.is_integer():
                    logging.warning('Non-uniform coarsening between level ' + str(lvl) + ' and ' + str(
                        lvl + 1) + '. Poorly tested.')
            else:
                self.m.append(1)
            self.setup_points_and_comm_info(lvl=lvl)
            self.step.append(problem[lvl].step)
            self.create_u(lvl=lvl)
            self.create_v_g(lvl=lvl)

        # Create coarse grid problem for direct solve
        self.create_coarsest_level()

        # Use or do not use nested iteration
        if nested_iteration:
            self.nested_iteration()

        # Stop timer for setup time
        self.comm_time.barrier()
        self.runtime_setup = time.time() - runtime_setup_start

        if self.output_fcn is not None and self.output_lvl == 2:
            self.output_fcn(self)

        self.log_info(f"Setup took {self.runtime_setup} s")

    def log_info(self, message: str) -> None:
        """
        Writes a message to the logger.
        Only one process

        :param message: Message
        """
        if self.comm_time_rank == 0:
            if self.spatial_parallel:
                if self.comm_space_rank == 0:
                    logging.info(message)
            else:
                logging.info(message)

    def iteration(self, lvl: int, cycle_type: str, iteration: int, first_f: bool) -> None:
        """
        MGRIT iteration on level lvl

        :param lvl: MGRIT level
        :param cycle_type: Cycle type
        :param iteration: Number of current iteration
        :param first_f: F-relaxation at the beginning
        """
        if lvl == self.lvl_max - 1:
            self.forward_solve(lvl=lvl)
            return

        if (lvl > 0 or (iteration == 0 and lvl == 0)) and first_f:
            self.f_relax(lvl=lvl)
            self.f_exchange(lvl=lvl)

        for _ in range(self.cf_iter):
            self.c_relax(lvl=lvl)
            self.c_exchange(lvl=lvl)
            self.f_relax(lvl=lvl)
            self.f_exchange(lvl=lvl)

        self.fas_residual(lvl=lvl)

        self.iteration(lvl=lvl + 1, cycle_type=cycle_type, iteration=iteration, first_f=True)

        self.error_correction(lvl=lvl)

        self.f_relax(lvl=lvl)

        if lvl != 0 and cycle_type == 'F':
            self.f_exchange(lvl=lvl)
            self.iteration(lvl=lvl, cycle_type='V', iteration=iteration, first_f=False)

    def f_relax(self, lvl: int) -> None:
        """
        F-relaxation on level lvl.

        F-relaxation updates solution values at F-points by propagating the solution
        from one C-point to all F-points up to the next C-point.

        :param lvl: MGRIT level
        """
        runtime_f = time.time()
        tmp_send = False
        req_s = None
        rank = self.comm_time_rank
        if self.index_local_f[lvl].size > 0:
            for i in np.nditer(self.index_local_f[lvl]):
                if self.comm_front[lvl] and i == np.min(self.index_local_f[lvl]):
                    self.u[lvl][0].unpack(self.comm_time.recv(source=self.get_from[lvl], tag=rank))
                if lvl == 0:
                    self.u[lvl][i] = self.step[lvl](u_start=self.u[lvl][i - 1],
                                                    t_start=self.t[lvl][i - 1],
                                                    t_stop=self.t[lvl][i])
                else:
                    self.u[lvl][i] = self.g[lvl][i] + self.step[lvl](u_start=self.u[lvl][i - 1],
                                                                     t_start=self.t[lvl][i - 1],
                                                                     t_stop=self.t[lvl][i])
                if self.comm_back[lvl] and i == np.max(self.index_local_f[lvl]):
                    tmp_send = True
                    req_s = self.comm_time.isend(self.u[lvl][-1].pack(), dest=self.send_to[lvl], tag=self.send_to[lvl])
        if tmp_send:
            req_s.wait()

        logging.debug(f"F-relax on {rank} took {time.time() - runtime_f} s")

    def c_relax(self, lvl: int) -> None:
        """
        C-relaxation on level lvl.

        C-relaxation updates solution values at C-points by propagating the
        solution from the preceeding F-points.

        :param lvl: MGRIT level
        """
        runtime_c = time.time()
        if self.index_local_c[lvl].size > 0:
            for i in np.nditer(self.index_local_c[lvl]):
                if i != 0 or self.comm_time_rank != 0:
                    if lvl == 0:
                        self.u[lvl][i] = self.step[lvl](u_start=self.u[lvl][i - 1],
                                                        t_start=self.t[lvl][i - 1],
                                                        t_stop=self.t[lvl][i])
                    else:
                        self.u[lvl][i] = self.g[lvl][i] + self.step[lvl](u_start=self.u[lvl][i - 1],
                                                                         t_start=self.t[lvl][i - 1],
                                                                         t_stop=self.t[lvl][i])
        logging.debug(f"C-relax on {self.comm_time_rank} took {time.time() - runtime_c} s")

    def convergence_criterion(self, iteration: int) -> None:
        """
        Stopping criterion based on the 2-norm of the space-time residual.

        Computes the space-time residual
          r_i = Phi(u_{i-1}) - u_i, i = 1, .... nt,
          r_0 = 0

        :param iteration: MGRIT iteration number
        """
        runtime_conv = time.time()
        r_norm = []

        self.f_exchange(lvl=0)
        self.c_exchange(lvl=0)

        if self.index_local_c[0].size > 0:
            for i in np.nditer(self.index_local_c[0]):
                if self.comm_time_rank != 0 or i != 0:
                    residual = self.step[0](u_start=self.u[0][i - 1], t_start=self.t[0][i - 1], t_stop=self.t[0][i]) - \
                               self.u[0][i]
                    r_norm.append(residual.norm())

        tmp = self.comm_time.allgather(r_norm)

        self.conv[iteration] = np.linalg.norm(np.array([item for sublist in tmp for item in sublist]), ord=self.t_norm)

        logging.debug(f"Convergence criterion on {self.comm_time_rank} took {time.time() - runtime_conv} s")

    def forward_solve(self, lvl: int) -> None:
        """
        Solves the problem directly on level lvl with time stepping.

        :param lvl: MGRIT level
        """

        runtime_fs = time.time()
        if self.lvl_max != 1:
            if self.comm_time_size != 1:
                tmp_g = self.comm_time.gather([self.g[lvl][i].pack() for i in self.index_local_c[lvl]], root=0)
                tmp_u = self.comm_time.gather([self.u[lvl][i].pack() for i in self.index_local_c[lvl]], root=0)
                if self.comm_time_rank == 0:
                    tmp_g = [item for sublist in tmp_g for item in sublist]
                    tmp_u = [item for sublist in tmp_u for item in sublist]
                    for i in range(len(self.g_coarsest)):
                        self.g_coarsest[i].unpack(tmp_g[i])
                        self.u_coarsest[i].unpack(tmp_u[i])
            else:
                self.g_coarsest = self.g[lvl]
                self.u_coarsest = self.u[lvl]

        if self.comm_time_rank == 0:
            for i in range(1, len(self.global_t[lvl])):
                self.u_coarsest[i] = self.g_coarsest[i] + self.step[lvl](u_start=self.u_coarsest[i - 1],
                                                                         t_start=self.global_t[lvl][i - 1],
                                                                         t_stop=self.global_t[lvl][i])

        if self.lvl_max != 1:
            if self.comm_time_size != 1:
                tmp_u_coarsest = self.comm_time.bcast([item.pack() for item in self.u_coarsest], root=0)
                for i in range(len(self.u_coarsest)):
                    self.u_coarsest[i].unpack(tmp_u_coarsest[i])

        if self.cpts[lvl].size > 0:
            self.u[lvl] = [self.u_coarsest[i] for i in self.cpts[lvl]]
            if self.comm_time_rank != 0:
                self.u[lvl] = [self.u[lvl][0]] + self.u[lvl]

        logging.debug(f"Forward solve on {self.comm_time_rank} took {time.time() - runtime_fs} s")

    def get_c_point(self, lvl: int) -> Application:
        """
        Exchanges the first/last C-point between two processes

        :param lvl: MGRIT level
        """
        rank = self.comm_time_rank
        tmp_send = False
        tmp = None
        req_s = None

        if self.send_to[lvl + 1] >= 0:
            req_s = self.comm_time.isend(self.u[lvl][self.index_local_c[lvl][-1]].pack(), dest=self.send_to[lvl + 1],
                                         tag=rank)
            tmp_send = True

        if self.get_from[lvl + 1] >= 0:
            tmp = self.problem[lvl].vector_template.clone_zero()
            tmp.unpack(self.comm_time.recv(source=self.get_from[lvl + 1], tag=self.get_from[lvl + 1]))

        if tmp_send:
            req_s.wait()
        return tmp

    def fas_residual(self, lvl: int) -> None:
        """
        Injects the fine-grid approximation and its residual
        from level lvl to the next coarser grid.

        :param lvl: MGRIT level
        """
        runtime_fas_res = time.time()
        tmp = self.get_c_point(lvl=lvl)
        rank = self.comm_time_rank

        if self.comm_time_rank != 0 and self.v[lvl + 1]:
            self.v[lvl + 1][0] = self.restriction[lvl](tmp)

        for i in range(len(self.index_local_c[lvl])):
            self.v[lvl + 1][i if rank == 0 else i + 1] = self.restriction[lvl](self.u[lvl][self.index_local_c[lvl][i]])

        self.u[lvl + 1] = [item.clone() for item in self.v[lvl + 1]]
        if np.size(self.index_local_c[lvl]) > 0:
            for i in range(len(self.index_local_c[lvl])):
                if i != 0 or self.comm_time_rank != 0:
                    if lvl == 0:
                        self.g[lvl + 1][self.index_local[lvl + 1][i]] = \
                            self.restriction[lvl](self.step[lvl](u_start=self.u[lvl][self.index_local_c[lvl][i] - 1],
                                                                 t_start=self.t[lvl][self.index_local_c[lvl][i] - 1],
                                                                 t_stop=self.t[lvl][self.index_local_c[lvl][i]])
                                                  - self.u[lvl][self.index_local_c[lvl][i]]) \
                            + self.v[lvl + 1][self.index_local[lvl + 1][i]] \
                            - self.step[lvl + 1](u_start=self.v[lvl + 1][self.index_local[lvl + 1][i] - 1],
                                                 t_start=self.t[lvl + 1][self.index_local[lvl + 1][i] - 1],
                                                 t_stop=self.t[lvl + 1][self.index_local[lvl + 1][i]])
                    else:
                        self.g[lvl + 1][self.index_local[lvl + 1][i]] = \
                            self.restriction[lvl](self.g[lvl][self.index_local_c[lvl][i]]
                                                  - self.u[lvl][self.index_local_c[lvl][i]]
                                                  + self.step[lvl](u_start=self.u[lvl][self.index_local_c[lvl][i] - 1],
                                                                   t_start=self.t[lvl][self.index_local_c[lvl][i] - 1],
                                                                   t_stop=self.t[lvl][self.index_local_c[lvl][i]])) \
                            + self.v[lvl + 1][self.index_local[lvl + 1][i]] \
                            - self.step[lvl + 1](u_start=self.v[lvl + 1][self.index_local[lvl + 1][i] - 1],
                                                 t_start=self.t[lvl + 1][self.index_local[lvl + 1][i] - 1],
                                                 t_stop=self.t[lvl + 1][self.index_local[lvl + 1][i]])

        logging.debug(f"Fas residual on {self.comm_time_rank} took {time.time() - runtime_fas_res} s")

    def nested_iteration(self) -> None:
        """
        Generates an initial approximation on the finest grid
        by solving the problem on the coarsest grid and interpolating
        the approximation to the finest level.
        """
        self.forward_solve(self.lvl_max - 1)

        for lvl in range(self.lvl_max - 2, -1, -1):
            for i in range(len(self.index_local[lvl + 1])):
                self.u[lvl][self.index_local_c[lvl][i]] = self.interpolation[lvl](
                    u=self.u[lvl + 1][self.index_local[lvl + 1][i]])

            self.f_exchange(lvl)
            self.c_exchange(lvl)
            if lvl > 0:
                self.iteration(lvl=lvl, cycle_type='V', iteration=0, first_f=True)

    def ouput_run_information(self) -> None:
        """
        Outputs information of pyMGRIT run.
        """
        msg = ['Run parameter overview',
               '  ' + '{0: <25}'.format(f'time interval') + ' : ' + '[' + str(self.problem[0].t[0]) + ', ' + str(
                   self.problem[0].t[-1]) + ']',
               '  ' + '{0: <25}'.format(f'number of time points ') + ' : ' + str(len(self.problem[0].t)),
               '  ' + '{0: <25}'.format(f'max dt ') + ' : ' + str(
                   np.max(self.problem[0].t[1:] - self.problem[0].t[:-1])),
               '  ' + '{0: <25}'.format(f'number of levels') + ' : ' + str(self.lvl_max),
               '  ' + '{0: <25}'.format(f'coarsening factors') + ' : ' + str(self.m[:-1]),
               '  ' + '{0: <25}'.format(f'cf_iter') + ' : ' + str(self.cf_iter),
               '  ' + '{0: <25}'.format(f'nested iteration') + ' : ' + str(self.nes_it),
               '  ' + '{0: <25}'.format(f'cycle type') + ' : ' + str(self.cycle_type),
               '  ' + '{0: <25}'.format(f'stopping tolerance') + ' : ' + str(self.tol),
               '  ' + '{0: <25}'.format(f'time communicator size') + ' : ' + str(self.comm_time_size),
               '  ' + '{0: <25}'.format(f'space communicator size') + ' : ' + str(self.comm_space_size)]
        self.log_info(message='\n'.join(msg))

    def f_exchange(self, lvl: int) -> None:
        """
        Point exchange on level lvl if the first point of a process is a C-point.

        Typically called after an F-point update.

        :param lvl: MGRIT level
        """
        runtime_ex = time.time()
        rank = self.comm_time_rank
        tmp_send = False
        req_s = None
        if self.last_is_f_point[lvl]:
            req_s = self.comm_time.isend(self.u[lvl][-1].pack(), dest=self.send_to[lvl], tag=self.send_to[lvl])
            tmp_send = True
        if self.first_is_c_point[lvl]:
            self.u[lvl][0].unpack(self.comm_time.recv(source=self.get_from[lvl], tag=rank))
        if tmp_send:
            req_s.wait()
        logging.debug(f"Exchange on {self.comm_time_rank} took {time.time() - runtime_ex} s")

    def c_exchange(self, lvl: int) -> None:
        """
        Point exchange on level lvl if the first point of a process is an F-point.

        Typically called after a C-point update.

        :param lvl: MGRIT level
        """
        runtime_ex = time.time()
        tmp_send = False
        req_s = None
        rank = self.comm_time_rank
        if self.last_is_c_point[lvl]:
            req_s = self.comm_time.isend(self.u[lvl][-1].pack(), dest=self.send_to[lvl], tag=self.send_to[lvl])
            tmp_send = True
        if self.first_is_f_point[lvl]:
            self.u[lvl][0].unpack(self.comm_time.recv(source=self.get_from[lvl], tag=rank))
        if tmp_send:
            req_s.wait()
        logging.debug(f"Exchange on {self.comm_time_rank} took {time.time() - runtime_ex} s")

    def solve(self) -> dict:
        """
        Driver function for solving the problem using MGRIT.

        Performs MGRIT iterations until a stopping criterion is fulfilled or
        the maximum number of iterations is reached.

        :return: dictionary with residual history, setup time, and solve time
        """
        # Start time of solve phase
        self.comm_time.barrier()
        self.log_info("Start solve")

        runtime_solve_start = time.time()
        for iteration in range(self.iter_max):

            self.solve_iter = iteration + 1
            time_it_start = time.time()
            self.iteration(lvl=0, cycle_type=self.cycle_type, iteration=iteration, first_f=True)
            self.comm_time.barrier()
            time_it_stop = time.time()
            self.convergence_criterion(iteration=iteration + 1)

            if iteration == 0:
                self.log_info('{0: <7}'.format(f"iter {iteration + 1}") +
                              '{0: <32}'.format(f" | conv: {self.conv[iteration + 1]}") +
                              '{0: <37}'.format(f" | conv factor: -") +
                              '{0: <35}'.format(f" | runtime: {time_it_stop - time_it_start} s"))
            else:
                self.log_info('{0: <7}'.format(f"iter {iteration + 1}") +
                              '{0: <32}'.format(f" | conv: {self.conv[iteration + 1]}") +
                              '{0: <37}'.format(f" | conv factor: {self.conv[iteration + 1] / self.conv[iteration]}") +
                              '{0: <35}'.format(f" | runtime: {time_it_stop - time_it_start} s"))

            if self.output_fcn is not None and self.output_lvl == 2:
                self.output_fcn(self)

            if self.conv[iteration + 1] < self.tol:
                break

        # Stop timer of solve phase
        self.comm_time.barrier()
        self.runtime_solve = time.time() - runtime_solve_start
        self.log_info(f"Solve took {self.runtime_solve} s")

        if self.output_fcn is not None and self.output_lvl == 1:
            self.output_fcn(self)

        self.ouput_run_information()
        return {'conv': self.conv[np.where(self.conv != 0)], 'time_setup': self.runtime_setup,
                'time_solve': self.runtime_solve}

    def error_correction(self, lvl: int) -> None:
        """
        Computes the error approximation on level lvl and
        updates the approximation on the next finer level.

        :param lvl: MGRIT level
        """
        for i in range(len(self.index_local_c[lvl])):
            error = self.interpolation[lvl](self.u[lvl + 1][self.index_local[lvl][i]] - self.v[lvl + 1][
                self.index_local[lvl][i]])
            self.u[lvl][self.index_local_c[lvl][i]] = self.u[lvl][self.index_local_c[lvl][i]] + error
        self.c_exchange(lvl=lvl)

    def split_points(self, length: int, size: int, rank: int) -> Tuple[int, int]:
        """
        Splits *length* points evenly in *size* parts and computes the index of the
        first point and block size of the local time interval.

        :param length: Number of points
        :param size: Number of processes
        :param rank: Process rank
        :return: Block size and index of first point
        """
        split = self.split_into(number_points=length, number_processes=size)

        return split[rank], np.sum(split[:rank]) if split[rank] > 0 else 0

    def setup_points_and_comm_info(self, lvl: int) -> None:
        """
        Computes local grid information for level *lvl*.
        Computes which process holds the previous and next point for each lvl and process.

        :param lvl: MGRIT level
        """
        self.global_t.append(np.copy(self.problem[lvl].t))
        points_time = np.size(self.global_t[lvl])
        all_pts = np.array(range(0, points_time))

        # Compute points per process
        # First level: Divide the points evenly
        # Other levels: Depends on the first level
        if lvl == 0:
            block_size_this_lvl, first_i_this_lvl = self.split_points(length=points_time,
                                                                      size=self.comm_time_size,
                                                                      rank=self.comm_time_rank)
            all_pts = all_pts[first_i_this_lvl:first_i_this_lvl + block_size_this_lvl]
            self.int_start = self.global_t[lvl][all_pts[0]]
            self.int_stop = self.global_t[lvl][all_pts[-1]]
        else:
            all_pts = np.where((self.global_t[lvl] >= self.int_start) & (self.global_t[lvl] <= self.int_stop))[0]

        # Compute C- and F-points
        if lvl != self.lvl_max - 1:
            all_cpts = np.where(np.in1d(self.problem[lvl].t, self.problem[lvl + 1].t))[0]
        else:
            all_cpts = np.array(range(0, points_time, self.m[lvl]))
        all_fpts = np.array(list(set(np.array(range(0, points_time))) - set(all_cpts)))
        cpts = np.sort(np.array(list(set(all_pts) - set(all_fpts)), dtype='int'))
        fpts = np.array(list(set(all_pts) - set(cpts)))
        fpts2 = np.array([item for sublist in np.array([np.array(xi) for xi in np.asarray(
            [list(map(itemgetter(1), g)) for k, g in groupby(enumerate(fpts), lambda x: x[0] - x[1])])])[::-1] for item
                          in
                          sublist])

        # Add ghost point if needed, set time interval
        with_ghost_point = False
        if self.comm_time_rank != 0 and all_pts.size > 0:
            with_ghost_point = True
            tmp = np.zeros(len(all_pts) + 1, dtype=int)
            tmp[0] = all_pts[0] - 1
            tmp[1:] = all_pts
            all_pts_with_ghost = tmp
        else:
            all_pts_with_ghost = all_pts

        self.t[lvl] = self.global_t[lvl][all_pts_with_ghost]
        self.cpts.append(cpts)

        # Communication in F-relax
        self.comm_front.append(bool(fpts.size > 0 and fpts[np.argmin(fpts)] - 1 in all_fpts))
        self.comm_back.append(bool(fpts.size > 0 and fpts[np.argmax(fpts)] + 1 in all_fpts))

        # Setup local indices
        self.index_local.append(np.nonzero(all_pts[:, None] == all_pts_with_ghost)[1])
        self.index_local_c.append(np.nonzero(cpts[:, None] == all_pts_with_ghost)[1])
        self.index_local_f.append(np.nonzero(fpts2[:, None] == all_pts_with_ghost)[1])

        # Communication before and after C- and F-relax
        self.first_is_c_point.append(
            bool(all_pts.size > 0 and all_pts[0] in cpts and all_pts[0] != 0 and all_pts[0] - 1 in all_fpts))
        self.first_is_f_point.append(bool(all_pts.size > 0 and all_pts[0] in fpts2 and all_pts[0] - 1 in all_cpts))
        self.last_is_c_point.append(bool(
            all_pts.size > 0 and all_pts[-1] in cpts and all_pts[-1] != points_time - 1 and all_pts[
                -1] + 1 in all_fpts))
        self.last_is_f_point.append(bool(
            all_pts.size > 0 and all_pts[-1] in fpts2 and all_pts[-1] != points_time - 1 and all_pts[
                -1] + 1 in all_cpts))

        # Setup communication info
        split = self.global_t[0][
            np.cumsum(self.split_into(number_points=len(self.global_t[0]), number_processes=self.comm_time_size)) - 1]
        tmp_send_to = -99
        tmp_get_from = -99
        if len(all_pts_with_ghost) > 0:
            if self.t[lvl][-1] != self.global_t[lvl][-1]:
                last_point_next = self.global_t[lvl][np.argwhere(self.global_t[lvl] == self.t[lvl][-1])[0][0] + 1]
                tmp_send_to = np.searchsorted(split, last_point_next)
            if with_ghost_point or self.t[lvl][0] != self.global_t[0][0]:
                tmp_get_from = np.searchsorted(split, self.t[lvl][0])
        self.send_to.append(tmp_send_to)
        self.get_from.append(tmp_get_from)

    def split_into(self, number_points: int, number_processes: int) -> np.ndarray:
        """
        Split points

        :param number_points: Number of points
        :param number_processes: Number of processes
        :return:
        """
        return np.array([int(number_points / number_processes + 1)] * (number_points % number_processes) +
                        [int(number_points / number_processes)] * (number_processes - number_points % number_processes))

    def create_u(self, lvl: int) -> None:
        """
        Creates solution vectors for all local time points on a given MGRIT level.

        :param lvl: MGRIT level
        """
        if lvl == 0:
            if self.random_init_guess:
                self.u.append([self.problem[lvl].vector_template.clone_rand() for _ in range(len(self.t[lvl]))])
            else:
                self.u.append([self.problem[lvl].vector_template.clone_zero() for _ in range(len(self.t[lvl]))])
        else:
            self.u.append([self.problem[lvl].vector_template.clone_zero() for _ in range(len(self.t[lvl]))])

        if self.comm_time_rank == 0:
            self.u[lvl][0] = self.problem[lvl].vector_t_start.clone()

    def create_v_g(self, lvl: int) -> None:
        """
        Creates vectors v and g for all local time points on a given MGRIT level.

        :param lvl: MGRIT level
        """
        if lvl == 0:
            self.v.append(None)
            self.g.append(None)
        else:
            self.v.append([item.clone_zero() for item in self.u[lvl]])
            self.g.append([item.clone_zero() for item in self.u[lvl]])

    def create_coarsest_level(self):
        """
        Creates vectors u and g for coarsest level
        """
        self.u_coarsest = [self.problem[-1].vector_template.clone_zero() for _ in range(len(self.global_t[-1]))]
        self.u_coarsest[0] = self.problem[-1].vector_t_start.clone()
        self.g_coarsest = [self.problem[-1].vector_template.clone_zero() for _ in range(len(self.global_t[-1]))]



"""
Helper functions for the induction
machine model "im_3kW". (https://gitlab.onelab.info/doc/models/-/wikis/Electric-machines)
"""

from typing import Tuple, List, Dict

import numpy as np
import scipy.spatial.qhull as qhull


def is_numeric(obj) -> bool:
    """
    Test if obj is numeric

    :param obj: test object
    :return: is numeric
    """
    try:
        obj + 0
        return True
    except TypeError:
        return False


def get_preresolution(file: str) -> int:
    """
    Read pre file and returns number of unknowns

    :param file: pre file
    :return: number of unknowns
    """
    with open(file) as f:
        content = f.readlines()
    ind = [idx for idx, s in enumerate(content) if '$DofData' in s][0]
    tmp = content[ind + 5].split()[-1]

    return int(tmp)


def set_resolution(file: str, t_start: float, u_start: np.ndarray, num_dofs: int) -> None:
    """
    Create resolution file

    :param file: file
    :param t_start: time associated with the input approximate solution u_start
    :param u_start: approximate solution for the input time t_start
    :param num_dofs: number of unknowns
    """
    dofpos = np.cumsum([0, num_dofs])
    com_str = ['$ResFormat /* GetDP 2.10.0, ascii */', '1.1 0', '$EndResFormat']

    for j in range(np.size(t_start)):
        for k in range(np.size(num_dofs)):
            com_str.append('$Solution  /* DofData #' + str(k) + ' */')
            com_str.append(str(k) + ' ' + str(t_start) + ' 0 ' + str(j))
            y = u_start[dofpos[k]: dofpos[k + 1]]
            com_str.append("\n".join(" ".join(map(str, line)) for line in np.vstack((np.real(y), np.imag(u_start))).T))
            com_str.append('$EndSolution\n')

    with open(file, "w") as fid:
        fid.write("\n".join(com_str))


def get_values_from(file: str) -> np.ndarray:
    """
    Read values from file

    :param file: result file
    :return: result value(s)
    """
    val = []
    with open(file) as fobj:
        for line in fobj:
            row = line.split()
            val.append(row[-1])
    return np.array(val, dtype=float)


def getdp_read_resolution(file: str, num_dofs: int) -> Tuple[np.ndarray, np.ndarray]:
    """
    Read unknown values from file

    :param file: result file
    :param num_dofs: number of unknowns
    :return: timepoint(s) and value(s)
    """
    # init solution vector, may contain several dofdata sets
    x = np.zeros((0, np.sum(num_dofs)))
    # init vector of time steps
    t = np.zeros(0)
    # init vector of time step numbers
    j = 0
    oldstep = 0
    # get positions of dofdata in x vector
    dofpos = np.cumsum([0, num_dofs])

    with open(file) as f:
        content = f.readlines()

    idx = 0
    while idx < len(content):
        if content[idx].find('$Solution') != -1:
            idx = idx + 1
            line = content[idx]
            idx = idx + 1
            tmp = line.split()
            tmp = [int(tmp[0]), float(tmp[1]), float(tmp[2]), int(tmp[3])]
            if oldstep < 1 + tmp[3]:
                j = j + 1
                oldstep = 1 + tmp[3]
                x = np.vstack((x, np.zeros((1, np.sum(num_dofs)))))
                t = np.hstack((t, 0))
            elif oldstep > 1 + tmp[3]:
                raise Exception('Exception reading file #s. time step #d is stored after #d', file,
                                tmp[3], oldstep - 1)
            k = 1 + tmp[0]
            t[j - 1] = tmp[1]
            # read complex dofdata set into solution vector
            xtmp = content[idx:idx + num_dofs]
            xtmp = np.array([list(map(float, s.split())) for s in xtmp])
            x[j - 1, dofpos[k - 1]:dofpos[k] + 1] = (xtmp[:, 0] + np.imag(xtmp[:, 1])).T
            idx = idx + num_dofs

        elif content[idx].find('$ResFormat') != -1:
            idx = idx + 1
            if not content[idx][0:3] == '1.1':
                raise Exception('Unknown file format version')
        else:
            idx = idx + 1

    if np.max(np.isnan(x)) or np.max(np.isnan(t)):
        raise Exception('getdp_read_resolution: file contains NaN | timepoint:', t)

    return t, x


def pre_file(file: str) -> Tuple[Dict, Dict, List]:
    """
    Read pre file and return mapping between nodes

    :param file: pre file
    :return: mapping between unknowns and grid points
    """
    with open(file) as f:
        content = f.readlines()

    mapping = content[9:-35]

    cor_to_un = {}
    un_to_cor = {}
    boundary = []

    for ma in mapping:
        row = ma.split()
        if row[4] != '0' and row[4] != '-1' and row[4] != '1':
            cor_to_un[row[1]] = row[4]
            un_to_cor[row[4]] = row[1]
        else:
            boundary = boundary + [row[1]]
    return cor_to_un, un_to_cor, boundary


def compute_data(pre: str, msh: str, new_unknown_start: int, inner_r: float = 0.04568666666666668) -> Dict:
    """
    Compute grid information

    :param pre: Pre file of mesh
    :param msh: Mesh
    :param new_unknown_start:
    :param inner_r: Radius machine
    :return:
    """
    cor_to_un, un_to_cor, boundary = pre_file(pre)
    nodes, nodes_r = get_nodes(msh)
    lines, elements, lines_r, elements_r = get_elements(msh)

    tmp = get_arrays(nodes, lines, elements, inner_r, un_to_cor, boundary, new_unknown_start)
    data = {'nodes': nodes, 'lines': lines, 'elements': elements, 'elementsR': elements_r, 'linesR': lines_r,
            'nodesR': nodes_r, 'corToUn': cor_to_un, 'unToCor': un_to_cor, 'boundary': boundary,
            'pointsCom': tmp['pointsCom'], 'pointsBou': tmp['pointsBou'], 'pointsInner': tmp['pointsInner'],
            'pointsBouInner': tmp['pointsBouInner'], 'elecom': tmp['elecom'], 'unknown': tmp['unknown'],
            'unknownCom': tmp['unknownCom'], 'indNodesToI': tmp['ind'],
            'boundaryNodes': tmp['boundaryNodes'], 'unknownComInner': tmp['unknownComInner'],
            'unknownComOuter': tmp['unknownComOuter'], 'unknownInner': tmp['unknownInner'],
            'unknownOuter': tmp['unknownOuter'], 'pointsOuter': tmp['pointsOuter'],
            'pointsBouOuter': tmp['pointsBouOuter'], 'mappingInnerToUnknown': tmp['mappingInnerToUnknown'],
            'mappingOuterToUnknown': tmp['mappingOuterToUnknown'], 'unknownNewInner': tmp['unknownNewInner'],
            'unknownNewOuter': tmp['unknownNewOuter'], 'mappingInnerToUnknownNew': tmp['mappingInnerToUnknownNew'],
            'mappingOuterToUnknownNew': tmp['mappingOuterToUnknownNew'], 'unknownNew': tmp['unknownNew']}
    return data

def check_version(msh_file:str):

    with open(msh_file) as f:
        content = f.readlines()

    if content[1].split()[0] != '4':
        raise Exception('Unsupported msh version. Required version: 4')

def compute_mesh_transfer(values: np.ndarray, vtx: np.ndarray, wts: np.ndarray, dif: int, dif2: int,
                          fill_value: float = np.nan) -> np.ndarray:
    """
    Compute mesh transfer

    :param values: vector to transform
    :param vtx: vertices
    :param wts: weights
    :param dif: difference for boundary conditions
    :param dif2: difference for boundary conditions
    :param fill_value: fill value
    :return: input values transfered to another grid
    """
    work = np.append(values, np.zeros(dif))
    ret = np.einsum('nj,nj->n', np.take(work, vtx), wts)
    ret[np.any(wts < 0, axis=1)] = fill_value
    ret = ret[:(np.size(ret) - dif2)]
    return ret


def get_nodes(file: str) -> Tuple[Dict, Dict]:
    """
    Get nodes from file

    :param file: Mesh file
    :return: Grid points
    """
    with open(file) as f:
        content = f.readlines()

    start = 0
    end = 0
    for i in range(len(content)):
        if content[i] == '$Nodes\n':
            start = i
        if content[i] == '$EndNodes\n':
            end = i

    nodes = content[start + 2:end]

    node_dict = {}
    point_to_node = {}
    for node in nodes:
        row = node.split()
        if len(row) > 1 and row[1] != '0' and row[1] != '1' and row[1] != '2':
            node_dict[row[0]] = np.array([float(row[1]), float(row[2])])
            point_to_node[row[1] + ' ' + row[2]] = row[0]

    return node_dict, point_to_node


def get_elements(file: str) -> Tuple[Dict, Dict, Dict, Dict]:
    """
    Get elements from file

    :param file: mesh file
    :return: elements
    """
    with open(file) as f:
        content = f.readlines()

    start_ele = 0
    end_ele = 0
    for i in range(len(content)):
        if content[i] == '$EndElements\n':
            end_ele = i
        if content[i] == '$Elements\n':
            start_ele = i

    ele = content[start_ele + 2:end_ele]
    ele_new_line = []
    ele_new_triangle = []
    i = 0
    while i < len(ele):
        num = int(ele[i].split()[-1])
        if len(ele[i + 1].split()) == 3:
            ele_new_line = ele_new_line + ele[i + 1:i + num + 1]
        elif len(ele[i + 1].split()) == 4:
            ele_new_triangle = ele_new_triangle + ele[i + 1:i + num + 1]
        i = i + num + 1

    ele_line_dict = {}
    ele_triangle_dict = {}
    ele_line_dict_reverse = {}
    ele_triangle_dict_reverse = {}

    for elem in ele_new_line:
        row = elem.split()
        ele_line_dict[row[0]] = np.array([row[1], row[2]])
        ele_line_dict_reverse[row[1] + ' ' + row[2]] = row[0]

    for elem in ele_new_triangle:
        row = elem.split()
        ele_triangle_dict[row[0]] = np.array([row[1], row[2], row[3]])
        ele_triangle_dict_reverse[row[1] + ' ' + row[2] + ' ' + row[3]] = row[0]

    return ele_line_dict, ele_triangle_dict, ele_line_dict_reverse, ele_triangle_dict_reverse


def cart2pol(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    """
    Transform Cartesian coordinates to polar

    :param x: x-coordinates
    :param y: y-coordinates
    :return: polar coordinates
    """
    return (x ** 2 + y ** 2) ** .5


def get_arrays(nodes_dict: Dict, lines_dict: Dict, elements_dict: Dict, inner_r: float,
               unknown_to_cor: Dict, boundary: List, new_start: int = 0) -> Dict:
    """
    Compute mapping between grids

    :param nodes_dict: grid points
    :param lines_dict: lines
    :param elements_dict: elements
    :param inner_r: radius machine
    :param unknown_to_cor: mapping between unknown and grid points
    :param boundary: boundary elements
    :param new_start: new unknowns
    :return: grid information
    """
    points_com = np.zeros((len(nodes_dict), 2))
    ind = {}

    i = 0
    for key, val in nodes_dict.items():
        points_com[i, 0] = val[0]
        points_com[i, 1] = val[1]
        ind[key] = i
        i = i + 1

    boundary_nodes = set([])
    for key, val in lines_dict.items():
        boundary_nodes.add(val[0])
        boundary_nodes.add(val[1])

    boundary_nodes = list(boundary_nodes)
    boundary_nodes.sort()

    i = 0
    points_bou = np.zeros((len(boundary_nodes), 2))
    for node in boundary_nodes:
        points_bou[i] = nodes_dict[node]
        i = i + 1

    elecom = np.zeros((len(elements_dict), 3), dtype=int)
    i = 0
    for key, val in elements_dict.items():
        elecom[i, 0] = ind[val[0]]
        elecom[i, 1] = ind[val[1]]
        elecom[i, 2] = ind[val[2]]
        i = i + 1

    unknown = np.zeros((len(unknown_to_cor), 2))
    unknown_com = np.zeros((len(unknown_to_cor) + len(boundary), 2))
    i = 0
    for key, val in unknown_to_cor.items():
        node = nodes_dict[val]
        unknown[i, 0] = node[0]
        unknown[i, 1] = node[1]
        unknown_com[i, 0] = node[0]
        unknown_com[i, 1] = node[1]
        i = i + 1
    for elem in boundary:
        node = nodes_dict[elem]
        unknown_com[i, 0] = node[0]
        unknown_com[i, 1] = node[1]
        i = i + 1

    unknown_new = np.copy(unknown[new_start:, :])

    r = cart2pol(points_com[:, 0], points_com[:, 1])
    inner = np.where(abs(r) - 1e-9 < abs(inner_r))[0]
    points_inner = points_com[inner]
    outer = np.where(abs(r) > abs(inner_r) - 1e-9)[0]
    points_outer = points_com[outer]

    r = cart2pol(points_bou[:, 0], points_bou[:, 1])
    inner_boundary_nodes = np.where(abs(r) - 1e-9 < abs(inner_r))[0]
    points_bou_inner = points_bou[inner_boundary_nodes]
    outer_boundary_nodes = np.where(abs(r) > abs(inner_r) + 1e-7)[0]
    points_bou_outer = points_bou[outer_boundary_nodes]

    r = cart2pol(unknown_com[:, 0], unknown_com[:, 1])
    inner = np.where(abs(r) - 1e-9 < abs(inner_r))[0]
    unknown_com_inner = unknown_com[inner]
    outer = np.where(abs(r) > abs(inner_r) + 1e-7)[0]
    unknown_com_outer = unknown_com[outer]

    r = cart2pol(unknown[:, 0], unknown[:, 1])
    inner = np.where(abs(r) - 1e-9 < abs(inner_r))[0]
    unknown_inner = unknown[inner]
    outer = np.where(abs(r) > abs(inner_r) + 1e-7)[0]
    unknown_outer = unknown[outer]

    r = cart2pol(unknown_new[:, 0], unknown_new[:, 1])
    inner = np.where(abs(r) - 1e-9 < abs(inner_r))[0]
    unknown_new_inner = unknown_new[inner]
    outer = np.where(abs(r) > abs(inner_r) + 1e-7)[0]
    unknown_new_outer = unknown_new[outer]

    mapping_inner_to_unknown = np.zeros(len(unknown_inner), dtype=int)
    mapping_outer_to_unknown = np.zeros(len(unknown_outer), dtype=int)

    mapping_inner_to_unknown_new = np.zeros(len(unknown_new_inner), dtype=int)
    mapping_outer_to_unknown_new = np.zeros(len(unknown_new_outer), dtype=int)

    k = 0
    s = 0
    for i in range(np.size(unknown_new, 0)):
        if unknown_new[i] in unknown_new_inner:
            mapping_inner_to_unknown_new[k] = i
            k = k + 1
        elif unknown_new[i] in unknown_new_outer:
            mapping_outer_to_unknown_new[s] = i
            s = s + 1

    k = 0
    i = 0
    s = 0
    for key, val in unknown_to_cor.items():
        node = nodes_dict[val]
        if node in unknown_inner:
            mapping_inner_to_unknown[k] = i
            k = k + 1
        if node in unknown_outer:
            mapping_outer_to_unknown[s] = i
            s = s + 1
        i = i + 1

    ret_dict = {
        'pointsCom': points_com,
        'pointsBou': points_bou,
        'pointsInner': points_inner,
        'pointsBouInner': points_bou_inner,
        'elecom': elecom,
        'unknown': unknown,
        'unknownCom': unknown_com,
        'ind': ind,
        'boundaryNodes': boundary_nodes,
        'pointsOuter': points_outer,
        'pointsBouOuter': points_bou_outer,
        'unknownComInner': unknown_com_inner,
        'unknownComOuter': unknown_com_outer,
        'unknownInner': unknown_inner,
        'unknownOuter': unknown_outer,
        'mappingInnerToUnknown': mapping_inner_to_unknown,
        'mappingOuterToUnknown': mapping_outer_to_unknown,
        'unknownNewInner': unknown_new_inner,
        'unknownNewOuter': unknown_new_outer,
        'mappingInnerToUnknownNew': mapping_inner_to_unknown_new,
        'mappingOuterToUnknownNew': mapping_outer_to_unknown_new,
        'unknownNew': unknown_new
    }
    return ret_dict


def interpolation_factors(data_coarse: Dict, data_fine: Dict) -> Dict:
    """
    Compute the interpolation factor for each point by two given grids

    :param data_coarse:
    :param data_fine:
    :return:
    """
    # vtxCom, wtsCom = interp_weights(data_coarse['unknownCom'], data_fine['unknown'][len(data_coarse['unToCor']):])
    vtx_inner, wts_inner = interp_weights(data_coarse['unknownComInner'], data_fine['unknownNewInner'])
    vtx_outer, wts_outer = interp_weights(data_coarse['unknownComOuter'], data_fine['unknownNewOuter'])

    add_bound_inner = np.size(data_coarse['unknownComInner'], 0) - np.size(data_coarse['unknownInner'], 0)
    add_bound_outer = np.size(data_coarse['unknownComOuter'], 0) - np.size(data_coarse['unknownOuter'], 0)
    size_lvl_stop = len(data_fine['corToUn'])
    size_lvl_start = len(data_coarse['corToUn'])
    mapping_inner = data_coarse['mappingInnerToUnknown']
    mapping_outer = data_coarse['mappingOuterToUnknown']
    mapping_inner_new = data_fine['mappingInnerToUnknownNew']
    mapping_outer_new = data_fine['mappingOuterToUnknownNew']

    ret_dict = {
        'vtxInner': vtx_inner,
        'wtsInner': wts_inner,
        'vtxOuter': vtx_outer,
        'wtsOuter': wts_outer,
        'addBoundInner': add_bound_inner,
        'addBoundOuter': add_bound_outer,
        'sizeLvlStop': size_lvl_stop,
        'sizeLvlStart': size_lvl_start,
        'mappingInner': mapping_inner,
        'mappingOuter': mapping_outer,
        'mappingInnerNew': mapping_inner_new,
        'mappingOuterNew': mapping_outer_new
    }

    return ret_dict


def interp_weights(xyz: np.ndarray, uvw: np.ndarray, d: int = 2, tol: float = 0.1) -> Tuple[np.ndarray, np.ndarray]:
    """
    Interpolation between two grids

    :param xyz: coarse grid points
    :param uvw: fine grid points
    :param d: dimensions
    :param tol: tolerance
    :return: vertices and weights
    """
    tri = qhull.Delaunay(xyz)
    simplex = tri.find_simplex(uvw, tol=tol)
    vertices = np.take(tri.simplices, simplex, axis=0)
    temp = np.take(tri.transform, simplex, axis=0)
    delta = uvw - temp[:, d]
    bary = np.einsum('njk,nk->nj', temp[:, :d, :], delta)
    wts = np.hstack((bary, 1 - bary.sum(axis=1, keepdims=True)))
    wts[wts < 0] = 0
    return vertices, wts



"""
Vector for the induction
machine model "im_3kW". (https://gitlab.onelab.info/doc/models/-/wikis/Electric-machines)
"""
from typing import List
import numpy as np

from pymgrit.core.vector import Vector


class VectorMachine(Vector):
    """
    Vector for the induction machine im_3kW
    """

    def __init__(self, u_front_size: int, u_middle_size: int, u_back_size: int) -> None:
        """
        Constructor
        :param u_front_size:
        :param u_middle_size:
        :param u_back_size:
        """
        super(VectorMachine, self).__init__()
        self.u_front_size = u_front_size
        self.u_middle_size = u_middle_size
        self.u_back_size = u_back_size
        self.u_front = np.zeros(u_front_size)
        self.u_back = np.zeros(u_back_size)
        self.u_middle = np.zeros(u_middle_size)
        self.jl = 0
        self.ua = 0
        self.ub = 0
        self.uc = 0
        self.ia = 0
        self.ib = 0
        self.ic = 0
        self.tr = 0

    def __add__(self, other: 'VectorMachine') -> 'VectorMachine':
        """
        Addition of two vector objects (self and other)

        :param other: vector object to be added to self
        :return: sum of vector object self and input object other
        """
        tmp = VectorMachine(len(self.u_front), len(self.u_middle), len(self.u_back))
        tmp.u_front = self.u_front + other.u_front
        tmp.u_back = self.u_back + other.u_back
        tmp.u_middle = self.u_middle + other.u_middle
        tmp.jl = self.jl + other.jl
        tmp.ia = self.ia + other.ia
        tmp.ib = self.ib + other.ib
        tmp.ic = self.ic + other.ic
        tmp.ua = self.ua + other.ua
        tmp.ub = self.ub + other.ub
        tmp.uc = self.uc + other.uc
        tmp.tr = self.tr + other.tr
        return tmp

    def __sub__(self, other: 'VectorMachine') -> 'VectorMachine':
        """
        Subtraction of two vector objects (self and other)

        :param other: vector object to be subtracted from self
        :return: difference of vector object self and input object other
        """
        tmp = VectorMachine(len(self.u_front), len(self.u_middle), len(self.u_back))
        tmp.u_front = self.u_front - other.u_front
        tmp.u_back = self.u_back - other.u_back
        tmp.u_middle = self.u_middle - other.u_middle
        tmp.jl = self.jl - other.jl
        tmp.ia = self.ia - other.ia
        tmp.ib = self.ib - other.ib
        tmp.ic = self.ic - other.ic
        tmp.ua = self.ua - other.ua
        tmp.ub = self.ub - other.ub
        tmp.uc = self.uc - other.uc
        tmp.tr = self.tr - other.tr
        return tmp

    def norm(self) -> float:
        """
        Norm of a vector object

        :return: 2-norm of vector object
        """
        tmp = np.append(np.append(self.u_front, self.u_middle), self.u_back)
        # tmp = np.append(tmp, [self.jl, self.ua, self.ub, self.uc, self.ia, self.ib, self.ic, self.tr])
        return np.linalg.norm(tmp)

    def clone(self):
        """
        Initial solution vector with all zeros
        :rtype: object
        """
        tmp = VectorMachine(len(self.u_front), len(self.u_middle), len(self.u_back))
        tmp.unpack(self.pack())
        return tmp

    def clone_zero(self) -> 'VectorMachine':
        """
        Initialize vector object with zeros

        :rtype: vector object with zero values
        """
        return VectorMachine(len(self.u_front), len(self.u_middle), len(self.u_back))

    def clone_rand(self) -> 'VectorMachine':
        """
        Initialize vector object with random values

        :rtype: vector object with random values
        """
        return VectorMachine(len(self.u_front), len(self.u_middle), len(self.u_back))

    def get_values(self) -> 'np.ndarray':
        """
        Get vector data

        :return: values of vector object
        """
        return np.append(np.append(self.u_front, self.u_middle), self.u_back)

    def set_values(self, values: np.ndarray, jl: float, ia: float, ib: float, ic: float, ua: float, ub: float,
                   uc: float, tr: float) -> None:
        """
        Set vector data
        """
        self.u_front = values[:self.u_front_size]
        self.u_middle = values[self.u_front_size:-self.u_back_size]
        self.u_back = values[-self.u_back_size:]
        self.jl = jl
        self.ia = ia
        self.ib = ib
        self.ic = ic
        self.ua = ua
        self.ub = ub
        self.uc = uc
        self.tr = tr

    def pack(self) -> List:
        """
        Pack data

        :return: values of vector object
        """
        send_obj = [self.u_front, self.u_middle, self.u_back, self.jl, self.ia, self.ib, self.ic, self.ua, self.ub,
                    self.uc, self.tr]
        return send_obj

    def unpack(self, values: List) -> None:
        """
        Unpack and set data

        :param values: values for vector object
        """
        self.u_front = values[0]
        self.u_middle = values[1]
        self.u_back = values[2]
        self.jl = values[3]
        self.ia = values[4]
        self.ib = values[5]
        self.ic = values[6]
        self.ua = values[7]
        self.ub = values[8]
        self.uc = values[9]
        self.tr = values[10]
